# pylint: disable=no-self-use,redefined-outer-name
"""
This module tests the :module:`xcc.connection` module.
"""

import json

import pytest
import requests
import responses
from requests.exceptions import HTTPError, RequestException
from responses import matchers

import xcc


class TestConnection:
    """Tests the :class:`xcc.Connection` class."""

    @pytest.mark.usefixtures("settings")
    def test_load_with_implicit_settings(self):
        """Tests that a connection can be loaded with an implicit settings configuration."""
        connection = xcc.Connection.load()
        assert connection.refresh_token == "j.w.t"
        assert connection.access_token is None
        assert connection.host == "example.com"
        assert connection.port == 80
        assert connection.tls is False

    @pytest.mark.usefixtures("settings")
    def test_load_with_explicit_settings(self):
        """Tests that a connection can be loaded with an explicit settings configuration."""
        settings = xcc.Settings(
            REFRESH_TOKEN=None,
            ACCESS_TOKEN="j.w.t",
            HOST="not.example.com",
            PORT=443,
            TLS=True,
        )

        connection = xcc.Connection.load(settings)
        assert connection.refresh_token is None
        assert connection.access_token == "j.w.t"
        assert connection.host == "not.example.com"
        assert connection.port == 443
        assert connection.tls is True

    @pytest.mark.usefixtures("settings")
    def test_load_with_keyword_arguments(self):
        """Tests that a connection can be loaded with new or overridden keyword arguments."""
        connection = xcc.Connection.load(host="not.example.com", headers={"User-Agent": "Bond"})
        assert connection.refresh_token == "j.w.t"
        assert connection.access_token is None
        assert connection.host == "not.example.com"
        assert connection.port == 80
        assert connection.tls is False
        assert connection.user_agent == "Bond"

    def test_missing_access_token_and_refresh_token(self):
        """Tests that a ValueError is raised when a connection is created
        without a refresh token or an access token.
        """
        match = (
            r"A refresh token \(e\.g\.\, Xanadu Cloud API key\) or an access "
            r"token must be provided to connect to the Xanadu Cloud"
        )
        with pytest.raises(ValueError, match=match):
            xcc.Connection()

    def test_access_token(self):
        """Tests that the correct access token is returned for a connection."""
        assert xcc.Connection(refresh_token="j.w.t").access_token is None
        assert xcc.Connection(access_token="j.w.t").access_token == "j.w.t"

    def test_refresh_token(self):
        """Tests that the correct refresh token is returned for a connection."""
        assert xcc.Connection(refresh_token="j.w.t").refresh_token == "j.w.t"
        assert xcc.Connection(access_token="j.w.t").refresh_token is None

    def test_tls(self, connection):
        """Tests that the correct TLS setting is returned for a connection."""
        assert connection.tls is True

    def test_scheme(self):
        """Tests that the correct scheme is returned for a connection."""
        assert xcc.Connection(refresh_token="j.w.t", tls=False).scheme == "http"
        assert xcc.Connection(refresh_token="j.w.t", tls=True).scheme == "https"

    def test_host(self, connection):
        """Tests that the correct host is returned for a connection."""
        assert connection.host == "test.xanadu.ai"

    def test_port(self, connection):
        """Tests that the correct port is returned for a connection."""
        assert connection.port == 443

    def test_api_version(self, connection):
        """Tests that the correct API version is returned for a connection."""
        assert connection.api_version == "0.4.0"
        connection._headers["Accept-Version"] = "1.2.3"  # pylint: disable=protected-access
        assert connection.api_version == "1.2.3"

    def test_user_agent(self, connection):
        """Tests that the correct user agent is returned for a connection."""
        assert connection.user_agent == f"XCC/{xcc.__version__} (API)"
        connection._headers["User-Agent"] = "Bond/1.2.3"  # pylint: disable=protected-access
        assert connection.user_agent == "Bond/1.2.3"

    def test_headers(self, connection):
        """Tests that the correct headers are returned for a connection."""
        connection._headers["Accept-Language"] = "en-CA"  # pylint: disable=protected-access
        assert connection.headers == {
            "Accept-Language": "en-CA",
            "Accept-Version": "0.4.0",
            "Authorization": "Bearer None",
            "User-Agent": f"XCC/{xcc.__version__} (API)",
        }

    def test_repr(self, connection):
        """Tests that the printable representation of a connection is correct."""
        assert repr(connection) == (
            "<Connection: refresh_token=j.w.t, access_token=None, url=https://test.xanadu.ai:443/>"
        )

    def test_url(self, connection):
        """Tests that the correct URL is derived for a connection path."""
        assert connection.url() == "https://test.xanadu.ai:443/"
        assert connection.url("/") == "https://test.xanadu.ai:443/"
        assert connection.url("/path/to/thing") == "https://test.xanadu.ai:443/path/to/thing"

    @responses.activate
    def test_ping_success(self, connection):
        """Tests that a ping succeeds when a connection is configured correctly."""
        responses.add(responses.GET, connection.url("healthz"), status=200)
        response = connection.ping()
        assert response.status_code == 200

    @responses.activate
    def test_ping_failure(self, connection):
        """Tests that a ping fails when a connection is configured incorrectly."""
        responses.add(responses.GET, connection.url("healthz"), status=400)
        with pytest.raises(HTTPError, match=r"400 Client Error: Bad Request"):
            connection.ping()

    @responses.activate
    def test_request_success_with_fresh_access_token(self, connection):
        """Tests that the correct response is returned for a connection with an
        access token that has not yet expired."""
        responses.add(responses.GET, connection.url("healthz"), status=200)
        assert connection.request("GET", "/healthz").status_code == 200
        assert len(responses.calls) == 1

    @responses.activate
    def test_request_success_with_expired_access_token(self, connection):
        """Tests that the correct response is returned for a connection with an
        expired access token."""
        responses.add(responses.GET, connection.url("healthz"), status=401)
        responses.add(responses.GET, connection.url("healthz"), status=200)
        responses.add(
            responses.POST,
            connection.url("auth/realms/platform/protocol/openid-connect/token"),
            status=200,
            body="{}",
        )

        assert connection.request("GET", "/healthz").status_code == 200
        assert len(responses.calls) == 3

    @responses.activate
    def test_request_failure_due_to_invalid_json(self, connection):
        """Tests that an HTTPError is raised when the status code of the HTTP
        response to a connection request indicates that an error has occurred
        and the body of the response is invalid JSON.
        """
        responses.add(responses.GET, connection.url("healthz"), status=400)
        with pytest.raises(HTTPError, match=r"400 Client Error: Bad Request"):
            connection.request("GET", "/healthz")

    @pytest.mark.parametrize(
        "meta, match",
        [
            (
                {
                    "_schema": ["Foo", "Bar"],
                    "size": ["Baz"],
                },
                r"Foo; Bar; Baz",
            ),
            (
                {
                    "head": {"0": ["Foo", "Bar"]},
                    "tail": {"0": ["Baz"], "1": ["Bud"]},
                },
                r"Foo; Bar; Baz; Bud",
            ),
            (
                {
                    "List": ["Foo"],
                    "Dict": {"0": ["Bar"]},
                },
                r"Foo; Bar",
            ),
        ],
    )
    @responses.activate
    def test_request_failure_due_to_validation_error(self, connection, meta, match):
        """Tests that an HTTPError is raised when the HTTP response of a
        connection request indicates that a validation error occurred and the
        body of the response contains a non-empty "meta" field.
        """
        body = {"code": "validation-error", "meta": meta}
        responses.add(responses.GET, connection.url("healthz"), status=400, body=json.dumps(body))

        with pytest.raises(HTTPError, match=match):
            connection.request("GET", "/healthz")

    @responses.activate
    def test_request_failure_due_to_detailed_error(self, connection):
        """Tests that an HTTPError is raised when the status code of the HTTP
        response to a connection request indicates that an error occurred and
        the body of the response contains a "detail" field.
        """
        body = {"code": "invalid-request", "detail": "The request body is empty."}
        responses.add(responses.GET, connection.url("healthz"), status=400, body=json.dumps(body))

        with pytest.raises(HTTPError, match=r"The request body is empty"):
            connection.request("GET", "/healthz")

    @responses.activate
    def test_request_failure_due_to_status_code(self, connection):
        """Tests that an HTTPError is raised when the status code of the HTTP
        response to a connection request indicates that an error occurred but
        no "detail" field is included in the JSON body of the response.
        """
        responses.add(responses.GET, connection.url("healthz"), status=403, body="{}")
        with pytest.raises(HTTPError, match=r"403 Client Error: Forbidden"):
            connection.request("GET", "/healthz")

    @responses.activate
    def test_request_failure_due_to_status_code_while_streaming(self, connection):
        """Tests that an HTTPError is raised when the status code of the HTTP
        response to a connection request where ``stream=True`` indicates that
        an error has occurred.
        """
        responses.add(responses.GET, connection.url("healthz"), status=403, body="{}")
        with pytest.raises(HTTPError, match=r"403 Client Error: Forbidden"):
            connection.request("GET", "/healthz", stream=True)

    def test_request_failure_due_to_timeout(self, monkeypatch, connection):
        """Tests that a RequestException is raised when a connection request times out."""

        def mock_request(*args, **kwargs):
            raise requests.exceptions.Timeout()

        monkeypatch.setattr("xcc.connection.requests.request", mock_request)

        match = r"GET request to 'https://test.xanadu.ai:443/healthz' timed out"
        with pytest.raises(RequestException, match=match):
            connection.request("GET", "/healthz")

    @responses.activate
    def test_request_failure_due_to_hostname_resolution(self, monkeypatch, connection):
        """Tests that a RequestException is raised when the hostname associated
        with a connection request cannot be resolved.
        """

        def mock_request(*args, **kwargs):
            raise requests.exceptions.ConnectionError("Name or service not known")

        monkeypatch.setattr("xcc.connection.requests.request", mock_request)

        with pytest.raises(RequestException, match=r"Failed to resolve hostname 'test.xanadu.ai'"):
            connection.request("GET", "/healthz")

    @responses.activate
    def test_request_headers(self, connection):
        """Tests that the correct headers are passed when the headers argument is not provided."""

        responses.add(
            url=connection.url("path"),
            method="POST",
            status=200,
            match=(matchers.header_matcher(connection.headers),),
        )

        connection.request(method="POST", path="path")

    @responses.activate
    @pytest.mark.parametrize("extra_headers", [{"X-Test": "data"}, {}])
    def test_request_extra_headers(self, connection, extra_headers):
        """Tests that the correct headers are passed when the headers argument is provided."""
        responses.add(
            url=connection.url("path"),
            method="POST",
            status=200,
            match=(matchers.header_matcher({**connection.headers, **extra_headers}),),
        )

        connection.request(method="POST", path="path", headers=extra_headers)

    @responses.activate
    def test_update_access_token_success(self, connection):
        """Tests that the access token of a connection can be updated."""
        responses.add(
            responses.POST,
            connection.url("auth/realms/platform/protocol/openid-connect/token"),
            status=200,
            body='{"access_token": "mock.access.token"}',
        )

        connection.update_access_token()

        assert connection.access_token == "mock.access.token"

    @responses.activate
    def test_update_access_token_success_with_invalid_json(self, connection):
        """Tests that a ValueError is raised when the HTTP response to an access
        token request has a 200 status code but the body of the response contains
        invalid JSON.
        """
        responses.add(
            responses.POST,
            connection.url("auth/realms/platform/protocol/openid-connect/token"),
            status=200,
        )

        match = r"Xanadu Cloud returned an invalid access token response\."
        with pytest.raises(ValueError, match=match):
            connection.update_access_token()

    @responses.activate
    def test_update_access_token_failure_due_to_invalid_json(self, connection):
        """Tests that an HTTPError is raised when the status code of an HTTP
        response to an access token request indicates that an error occurred and
        the body of the response contains invalid JSON.
        """
        responses.add(
            responses.POST,
            connection.url("auth/realms/platform/protocol/openid-connect/token"),
            status=400,
        )

        with pytest.raises(HTTPError, match=r"400 Client Error: Bad Request"):
            connection.update_access_token()

    @responses.activate
    def test_update_access_token_failure_due_to_invalid_refresh_token(self, connection):
        """Tests that an HTTPError is raised when the HTTP response to an access
        token request indicates that the refresh token (API key) is invalid.
        """
        responses.add(
            responses.POST,
            connection.url("auth/realms/platform/protocol/openid-connect/token"),
            status=400,
            body='{"error": "invalid_grant"}',
        )

        match = r"Refresh token \(e\.g\.\, Xanadu Cloud API key\) is invalid"
        with pytest.raises(HTTPError, match=match) as excinfo:
            connection.update_access_token()
        assert hasattr(excinfo.value, "response")
        assert excinfo.value.response.status_code == 400
        assert excinfo.value.response.json()["error"] == "invalid_grant"

    @responses.activate
    def test_update_access_token_failure_due_to_status_code(self, connection):
        """Tests that an HTTPError is raised when the HTTP response to an access
        token request contains valid JSON but its status code indicates that an
        error occurred.
        """
        responses.add(
            responses.POST,
            connection.url("auth/realms/platform/protocol/openid-connect/token"),
            status=403,
            body="{}",
        )

        with pytest.raises(HTTPError, match=r"403 Client Error: Forbidden"):
            connection.update_access_token()



# pylint: disable=no-self-use,redefined-outer-name
"""
This module tests the :module:`xcc.commands` module.
"""

import json
from inspect import cleandoc

import numpy as np
import pytest
from fire.core import FireError

import xcc
import xcc.commands
from xcc.util import cached_property


class MockDevice(xcc.Device):
    """Mock :class:`xcc.Device` class with an offline implementation of each
    property and function which is directly accessed by a CLI command.
    """

    @staticmethod
    def list(connection, status=None):
        connection = xcc.commands.load_connection()
        return [MockDevice(target, connection) for target in ("foo", "bar")]

    @property
    def overview(self):
        return {"target": self.target}

    @property
    def status(self):
        return "online"

    @cached_property
    def certificate(self):
        return {"temperature": "300 Kelvin"}

    @cached_property
    def specification(self):
        return {"modes": 42}


class MockJob(xcc.Job):
    """Mock :class:`xcc.Job` class with an offline implementation of each
    property and function which is directly accessed by a CLI command.
    """

    @staticmethod
    def list(connection, limit=10, ids=None, status=None):
        connection = xcc.commands.load_connection()
        return [MockJob(id_, connection) for id_ in ("foo", "bar", "baz")][:limit]

    @staticmethod
    def submit(connection, name, target, circuit, language="blackbird:1.0"):
        return MockJob(id_=name, connection=connection)

    @property
    def overview(self):
        return {"id": self.id}

    @property
    def status(self):
        return "cancelled"

    @property
    def circuit(self):
        return "MeasureFock() | [0, 1, 2, 3]"

    @property
    def language(self):
        return "blackbird:1.0"

    def get_result(self, integer_overflow_protection=True):
        return {"output": [np.zeros((4, 4))], "metadata": np.ones((2, 2))}

    def cancel(self):
        pass


@pytest.fixture(autouse=True)
def device(monkeypatch) -> None:
    """Replaces the :class:`xcc.Device` class with the :class:`MockDevice` class in the CLI."""
    monkeypatch.setattr("xcc.commands.Device", MockDevice)


@pytest.fixture(autouse=True)
def job(monkeypatch) -> None:
    """Replaces the :class:`xcc.Job` class with the :class:`MockJob` class in the CLI."""
    monkeypatch.setattr("xcc.commands.Job", MockJob)


@pytest.mark.usefixtures("settings")
def test_load_connection(monkeypatch):
    """Tests that a connection can be loaded."""
    monkeypatch.setattr("xcc.commands.__version__", "1.2.3-alpha")

    connection = xcc.commands.load_connection()
    assert connection.refresh_token == "j.w.t"
    assert connection.host == "example.com"
    assert connection.port == 80
    assert connection.tls is False
    assert connection.user_agent == "XCC/1.2.3-alpha (CLI)"


# Settings CLI
# ------------------------------------------------------------------------------


class TestGetSetting:
    """Tests the :func:`xcc.commands.get_setting()` CLI command."""

    def test_valid(self):
        """Tests that the value of a setting can be retrieved."""
        assert xcc.commands.get_setting(name="PORT") == 80

    def test_mixed_case(self):
        """Tests that the name of a setting is case-insensitive."""
        assert xcc.commands.get_setting(name="PoRt") == 80

    def test_invalid_name(self):
        """Tests that a ValueError is raised when the name of a setting is invalid."""
        with pytest.raises(ValueError, match=r"The setting name 'DNE' must be one of \[.*\]"):
            xcc.commands.get_setting(name="DNE")

    def test_missing_value(self, settings):
        """Tests that a ValueError is raised when the value of a setting is ``None``."""
        settings.ACCESS_TOKEN = None
        settings.save()

        match = r"The ACCESS_TOKEN setting does not currently have a value"
        with pytest.raises(ValueError, match=match):
            xcc.commands.get_setting(name="ACCESS_TOKEN")


def test_list_settings():
    """Tests that the :func:`xcc.commands.list_settings()` CLI command lists settings."""
    have_settings = json.loads(xcc.commands.list_settings())
    want_settings = {
        "REFRESH_TOKEN": "j.w.t",
        "ACCESS_TOKEN": None,
        "HOST": "example.com",
        "PORT": 80,
        "TLS": False,
    }
    assert have_settings == want_settings


class TestSetSetting:
    """Tests the :func:`xcc.commands.set_setting()` CLI command."""

    def test_valid(self):
        """Tests that the value of a setting can be updated."""
        xcc.commands.set_setting(name="PORT", value=123)
        assert xcc.Settings().PORT == 123

    def test_mixed_case(self):
        """Tests that the name of a setting is case-insensitive."""
        xcc.commands.set_setting(name="PoRt", value=123)
        assert xcc.Settings().PORT == 123

    @pytest.mark.parametrize(
        "name, value, message",
        [
            ("TLS", True, "Successfully updated TLS setting to True."),
            ("PORT", 123, "Successfully updated PORT setting to 123."),
            ("HOST", "foo", "Successfully updated HOST setting to 'foo'."),
        ],
    )
    def test_message(self, name, value, message):
        """Tests that the correct message is displayed when a setting is updated."""
        assert xcc.commands.set_setting(name=name, value=value) == message

    def test_invalid_name(self):
        """Tests that a ValueError is raised when the name of a setting is invalid."""
        with pytest.raises(ValueError, match=r"The setting name 'DNE' must be one of \[.*\]\."):
            xcc.commands.set_setting(name="DNE", value="N/A")

    def test_invalid_value(self):
        """Tests that a ValueError is raised when the value of a setting is invalid."""
        match = r"Failed to update PORT setting: value is not a valid integer"
        with pytest.raises(ValueError, match=match):
            xcc.commands.set_setting(name="PORT", value="string")


# Device CLI
# ------------------------------------------------------------------------------


class TestGetDevice:
    """Tests the :func:`xcc.commands.get_device()` CLI command."""

    def test_overview(self):
        """Tests that the overview of a device can be retrieved."""
        have_overview = json.loads(xcc.commands.get_device(target="foo"))
        want_overview = {"target": "foo"}
        assert have_overview == want_overview

    def test_status(self):
        """Tests that the status of a device can be retrieved."""
        assert xcc.commands.get_device(target="foo", status=True) == "online"

    def test_certificate(self):
        """Tests that the certificate of a device can be retrieved."""
        have_certificate = json.loads(xcc.commands.get_device(target="foo", certificate=True))
        want_certificate = {"temperature": "300 Kelvin"}
        assert have_certificate == want_certificate

    def test_specification(self):
        """Tests that the specification of a device can be retrieved."""
        have_specification = json.loads(xcc.commands.get_device(target="foo", specification=True))
        want_specification = {"modes": 42}
        assert have_specification == want_specification

    def test_invalid_number_of_flags(self):
        """Tests that a FireError is raised when more than one flag is specified."""
        with pytest.raises(FireError, match=r"At most one device property can be selected"):
            xcc.commands.get_device(target="foo", status=True, certificate=True)


def test_list_devices():
    """Tests that the :func:`xcc.commands.list_devices()` CLI command lists devices."""
    have_list = json.loads(xcc.commands.list_devices())
    want_list = [{"target": "foo"}, {"target": "bar"}]
    assert have_list == want_list


# Job CLI
# ------------------------------------------------------------------------------


class TestCancelJob:
    """Tests the :func:`xcc.commands.cancel_job()` CLI command."""

    def test_cancel_success(self):
        """Tests that the correct message is displayed when a job was cancelled."""
        have_message = xcc.commands.cancel_job(id="foo")
        want_message = "Successfully cancelled job with ID 'foo'."
        assert have_message == want_message

    def test_cancel_failure(self, monkeypatch):
        """Tests that a ValueError is raised if a job could not be cancelled."""
        monkeypatch.setattr("xcc.commands.Job.status", "completed")
        with pytest.raises(ValueError, match=r"Job with ID 'foo' was not cancelled in time\."):
            xcc.commands.cancel_job(id="foo")


class TestGetJob:
    """Tests the :func:`xcc.commands.get_job()` CLI command."""

    def test_overview(self):
        """Tests that the overview of a job can be retrieved."""
        have_overview = json.loads(xcc.commands.get_job(id="foo"))
        want_overview = {"id": "foo"}
        assert have_overview == want_overview

    def test_status(self):
        """Tests that the status of a job can be retrieved."""
        assert xcc.commands.get_job(id="foo", status=True) == "cancelled"

    def test_circuit(self):
        """Tests that the circuit of a job can be retrieved."""
        assert xcc.commands.get_job(id="foo", circuit=True) == "MeasureFock() | [0, 1, 2, 3]"

    def test_result(self):
        """Tests that the result of a job can be retrieved."""
        have_result = xcc.commands.get_job(id="foo", result=True)
        want_result = cleandoc(
            """
            {
                "metadata": [[1.0, 1.0], [1.0, 1.0]],
                "output": [[[0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0]]]
            }
            """
        )
        assert have_result == want_result

    def test_invalid_number_of_flags(self):
        """Tests that a FireError is raised when more than one flag is specified."""
        with pytest.raises(FireError, match=r"At most one job property can be selected"):
            xcc.commands.get_job(id="foo", status=True, result=True)


def test_list_jobs():
    """Tests that the :func:`xcc.commands.list_jobs()` CLI command lists jobs."""
    have_list = json.loads(xcc.commands.list_jobs(limit=2))
    want_list = [{"id": "foo"}, {"id": "bar"}]
    assert have_list == want_list


def test_submit_job():
    """Tests that a job can be submitted using the :func:`xcc.commands.submit_job()` CLI command."""
    have_overview = json.loads(xcc.commands.submit_job(name="foo", target="bar", circuit="baz"))
    want_overview = {"id": "foo"}
    assert have_overview == want_overview


# Other CLI
# ------------------------------------------------------------------------------


def test_ping(monkeypatch):
    """Tests that the output of the :func:`xcc.commands.ping()` command
    is correct when the connection to the Xanadu Cloud is successful.
    """
    monkeypatch.setattr("xcc.commands.Connection.ping", lambda _: None)
    assert xcc.commands.ping() == "Successfully connected to the Xanadu Cloud."


def test_version(monkeypatch):
    """Tests that the output of the :func:`xcc.commands.version()` command
    has the correct form.
    """
    monkeypatch.setattr("xcc.commands.__version__", "1.2.3-alpha")
    assert xcc.commands.version() == "Xanadu Cloud Client version 1.2.3-alpha"



"""
This module contains the :class:`~xcc.Job` class.
"""

from __future__ import annotations

import io
import time
from datetime import datetime, timedelta
from itertools import count, takewhile
from typing import Any, Collection, List, Mapping, Optional, Sequence, Union

import dateutil.parser
import numpy as np

from .connection import Connection
from .util import cached_property


class Job:
    """Represents a job on the Xanadu Cloud.

    Args:
        id_ (str): ID of the job
        connection (Connection): connection to the Xanadu Cloud

    .. note::

        For performance reasons, the properties of a job are lazily fetched and
        stored in a cache. This cache can be cleared at any time by calling
        :meth:`Job.clear`.

    .. warning::

        The :class:`xcc.Job` class transparently contacts the Xanadu Cloud when
        an uncached job property is accessed. This means that requesting a job
        property for the first time may take longer than expected.

    **Example:**

    The following example shows how to use the :class:`Job` class to submit and
    track a job for a simulator on the Xanadu Cloud. First, a connection is
    established to the Xanadu Cloud:

    >>> import xcc
    >>> connection = xcc.Connection.load()

    Next, the parameters of the desired job are prepared. At present, this
    includes the name, target, circuit, and language of the job:

    >>> import inspect
    >>> name = "example"
    >>> target = "simulon_gaussian"
    >>> circuit = inspect.cleandoc(
            f\"\"\"
            name {name}
            version 1.0
            target {target} (shots=3)
            MeasureFock() | [0, 1, 2, 3]
            \"\"\"
        )
    >>> language = "blackbird:1.0"

    The job is then submitted to the Xanadu Cloud using the :func:`Job.submit`
    function:

    >>> job = xcc.Job.submit(
            connection,
            name=name,
            target=target,
            circuit=circuit,
            language=language,
        )

    At this point, the Xanadu Cloud has received the job; however, it may take
    some time before the result of the job is available. To wait until the job
    is finished, a blocking call is made to :meth:`Job.wait`:

    >>> job.wait()

    Finally, the status, result, running time, etc. of the job are retrieved by
    accessing the corresponding properties of the job:

    >>> job.status
    'complete'
    >>> job.result
    {'output': [array([[0, 0, 0, 0],
       [0, 0, 0, 0],
       [0, 0, 0, 0]])]}
    >>> job.running_time
    datetime.timedelta(microseconds=123456)
    """

    @staticmethod
    def list(
        connection: Connection,
        limit: int = 5,
        ids: Optional[Collection[str]] = None,
        status: Optional[str] = None,
    ) -> Sequence[Job]:
        """Returns jobs submitted to the Xanadu Cloud.


        Args:
            connection (Connection): connection to the Xanadu Cloud
            limit (int): maximum number of jobs to retrieve
            ids (Collection[str], optional): IDs of the jobs to retrieve; if at
                least one ID is specified, ``limit`` will be set to the length
                of the ID collection
            status (str, optional): filter jobs by the given status (if not ``None``)

        Returns:
            Sequence[Job]: jobs which were submitted on the Xanadu Cloud by the
            user associated with the Xanadu Cloud connection
        """
        size = len(ids) if ids else limit

        params = {"size": size, "id": ids, "status": status}
        response = connection.request("GET", "/jobs", params=params)

        jobs = []

        for details in response.json()["data"]:
            job = Job(details["id"], connection=connection)
            job._details = details  # pylint: disable=protected-access
            jobs.append(job)

        return jobs

    @staticmethod
    def submit(
        connection: Connection,
        name: Optional[str],
        target: str,
        circuit: str,
        language: str,
    ) -> Job:
        """Submits a job to the Xanadu Cloud.

        Args:
            connection (Connection): connection to the Xanadu Cloud
            name (str, optional): name of the job
            target (str): target of the job
            circuit (str): circuit of the job
            language (str): language of the job

        Returns:
            Job: job submitted to the Xanadu Cloud
        """
        payload = {
            "name": name,
            "target": target,
            "circuit": circuit,
            "language": language,
        }
        response = connection.request("POST", "/jobs", json=payload)
        details = response.json()

        job = Job(details["id"], connection=connection)

        # pylint: disable=protected-access
        job._details = details
        job._circuit = {"circuit": circuit}

        return job

    def __init__(self, id_: str, connection: Connection) -> None:
        self._id = id_
        self._connection = connection

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Returns the ID of a job."""
        return self._id

    @property
    def overview(self) -> Mapping[str, Any]:
        """Returns an overview of a job.

        Returns:
            Mapping[str, Any]: mapping from field names to values for this job
            as determined by the needs of a Xanadu Cloud user.
        """
        return {
            "id": self.id,
            "name": self.name,
            "status": self.status,
            "target": self.target,
            "language": self.language,
            "created_at": self.created_at,
            "finished_at": self.finished_at,
            "running_time": self.running_time,
            "metadata": self.metadata,
        }

    @cached_property
    def result(self) -> Mapping[str, Union[np.ndarray, List[np.ndarray]]]:
        """Returns the result of a job.

        Returns:
            Mapping[str, Union[np.ndarray, List[np.ndarray]]]: The result of
            this job.

        .. seealso::

            Implemented in terms of :meth:`Job.get_result`.

        .. note::

            NumPy integers will be converted to ``np.int64`` objects to
            facilitate safer post-processing.
        """
        return self.get_result(integer_overflow_protection=True)

    @property
    def created_at(self) -> datetime:
        """Returns when a job was created.

        Returns:
            datetime: datetime when this job was created
        """
        return dateutil.parser.isoparse(self._details["created_at"])

    @property
    def started_at(self) -> Optional[datetime]:
        """Returns when a job started.

        Returns:
            datetime, optional: datetime when this job started
        """
        datetime_ = self._details["started_at"]
        return None if datetime_ is None else dateutil.parser.isoparse(datetime_)

    @property
    def finished_at(self) -> Optional[datetime]:
        """Returns when a job finished.

        Returns:
            datetime, optional: datetime when this job finished
        """
        datetime_ = self._details["finished_at"]
        return None if datetime_ is None else dateutil.parser.isoparse(datetime_)

    @property
    def running_time(self) -> Optional[timedelta]:
        """Returns the running time of a job.

        Returns:
            timedelta, optional: running time of this job
        """
        seconds = self._details["running_time"]
        return None if seconds is None else timedelta(seconds=seconds)

    @property
    def circuit(self) -> str:
        """Returns the circuit of a job.

        Returns:
            str: circuit of this job
        """
        return self._circuit["circuit"]

    @property
    def language(self) -> str:
        """Returns the language of a job.

        Returns:
            str: language of this job
        """
        return self._details["language"]

    @property
    def name(self) -> Optional[str]:
        """Returns the name of a job.

        Returns:
            str, optional: name of this job
        """
        return self._details["name"]

    @property
    def target(self) -> str:
        """Returns the target of a job.

        Returns:
            str: target of this job
        """
        return self._details["target"]

    @property
    def status(self) -> str:
        """Returns the current status of a job.

        Returns:
            str: status of this job ("open", "queued", "cancelled", "failed",
            "cancel_pending", or "complete")
        """
        return self._details["status"]

    @property
    def finished(self) -> bool:
        """Returns whether this job has finished.

        Returns:
            bool: ``True`` iff the job status is "cancelled", "complete", or
            "failed"
        """
        return self.status in ("cancelled", "complete", "failed")

    @property
    def metadata(self) -> Mapping[str, Any]:
        """Returns the metadata of a job.

        Returns:
            Mapping[str, Any]: metadata of this job

        .. note::

            Error details for failed jobs are reported here.
        """
        return self._details["meta"]

    @cached_property
    def _details(self) -> Mapping[str, Any]:
        """Returns the details of a job.

        Returns:
            Mapping[str, Any]: mapping from field names to values for this job
                as determined by the Xanadu Cloud job endpoint.

        .. note::

            These fields are not intended to be directly accessed by external
            callers. Instead, they should be individually retrieved through
            their associated public properties.
        """
        return self._connection.request("GET", f"/jobs/{self.id}").json()

    @cached_property
    def _circuit(self) -> Mapping[str, str]:
        """Returns the circuit of a job.

        Returns:
            Mapping[str, str]: mapping with a "circuit" field

        .. note::

            The circuit should be retrieved through the :attr:`Device.circuit` property.
        """
        return self._connection.request("GET", f"/jobs/{self.id}/circuit").json()

    def __repr__(self) -> str:
        """Returns a printable representation of a job."""
        return f"<{self.__class__.__name__}: id={self.id}>"

    def get_result(
        self,
        integer_overflow_protection: bool = True,
    ) -> Mapping[str, Union[np.ndarray, List[np.ndarray]]]:
        """Returns the result of a job.

        Args:
            integer_overflow_protection (bool): convert all NumPy integers into
                64-bit NumPy integers during post-processing; setting this
                option to ``False`` can significantly reduce memory consumption
                at the risk of introducing bugs during e.g. cubing operations

        Returns:
            Mapping[str, Union[np.ndarray, List[np.ndarray]]]: The result of
            this job. Each job result has an "output" key associated with a list
            of NumPy arrays representing the output of the job; all other keys
            represent metadata related to the interpretation of the job output.

        Raises:
            TypeError: if the job result is not stored in the .npz file format

        .. warning::

            The value returned by this method is not cached.
        """
        # Streaming the response in chunks is necessary to fetch large results.
        response = self._connection.request("GET", f"/jobs/{self.id}/result", stream=True)

        with io.BytesIO() as buffer:
            # The chunk size below (i.e., 2^15 bytes) is less than the maximum
            # size of a TCP packet (i.e., 2^16 bytes) but is otherwise arbitrary.
            for chunk in response.iter_content(chunk_size=32768):
                buffer.write(chunk)

            # Seeking to 0 prepares the buffer for reading.
            buffer.seek(0)

            result = np.load(buffer, allow_pickle=False)

            if not isinstance(result, np.lib.npyio.NpzFile):
                raise TypeError("Job result is not an .npz file")

            result = dict(result)

        if integer_overflow_protection:
            for key, array in result.items():
                if np.issubdtype(array.dtype, np.integer):
                    result[key] = array.astype(np.int64)

        # For convenience, move all positional arguments (i.e., those that have
        # an "arr_" prefix) to a separate key named "output".
        result["output"] = []
        for i in takewhile(lambda i: f"arr_{i}" in result, count()):
            array = result.pop(f"arr_{i}")
            result["output"].append(array)

        return result

    def cancel(self) -> None:
        """Cancels a job."""
        if not self.finished:
            self._connection.request("PATCH", f"/jobs/{self.id}", json={"status": "cancelled"})
            self.clear()

    def clear(self) -> None:
        """Clears the details, circuit, and result caches of a job."""
        del self._details
        del self._circuit
        del self.result

    def wait(self, delay: float = 1) -> None:
        """Waits for a job to finish.

        Args:
            delay (float): number of seconds to wait between polling requests
        """
        while not self.finished:
            time.sleep(delay)
            self.clear()



"""
This module contains the :class:`~xcc.Settings` class and related helper functions.
"""

import os
import re
from typing import Optional

from appdirs import user_config_dir
from dotenv import dotenv_values, set_key, unset_key
from pydantic import BaseSettings

# Matches when string contains chars outside Base64URL set
# https://base64.guru/standards/base64url
# https://en.wikipedia.org/wiki/Base64#Variants_summary_table
# Also, the dot '.' to separate sections.
_BASE64URLRE = re.compile(r"[^\.A-Za-z0-9_-]+")


def get_path_to_env_file() -> str:
    """Returns the path to the .env file containing the Xanadu Cloud connection settings."""
    path_to_config_dir = user_config_dir(appname="xanadu-cloud", appauthor="Xanadu")
    return os.path.join(path_to_config_dir, ".env")


def get_name_of_env_var(key: str = "") -> str:
    """Returns the name of the Xanadu Cloud environment variable associated with the given key."""
    return f"XANADU_CLOUD_{key}"


def _check_for_invalid_values(key: str, val: str) -> None:
    """
    Check for conditions that make saving the env_file
    dangerous to the user.

    - REFRESH_TOKEN must not contain characters outside Base64URL set

    Args:
        key (str): .env file key
        val (str): .env file value

    Raises:
        ValueError: if the value should not be saved to the .env file

    """

    if key == "REFRESH_TOKEN" and val is not None and re.search(_BASE64URLRE, val) is not None:
        raise ValueError("REFRESH_TOKEN contains non-JWT character(s)")


class Settings(BaseSettings):
    """Represents the configuration for connecting to the Xanadu Cloud.

    The location where this configuration is saved depends on the current
    operating system. Specifically,

    * Windows: ``C:\\Users\\%USERNAME%\\AppData\\Local\\Xanadu\\xanadu-cloud\\.env``

    * MacOS: ``/home/$USER/Library/Application\\ Support/xanadu-cloud/.env``

    * Linux: ``/home/$USER/.config/xanadu-cloud/.env``

    **Example:**

    The following example shows how to use the :class:`Settings` class to load
    and save a Xanadu Cloud configuration. To begin, loading a configuration is
    as simple as instantiating a settings object:

    >>> import xcc
    >>> settings = xcc.Settings()
    >>> settings
    REFRESH_TOKEN=None ACCESS_TOKEN=None HOST='platform.xanadu.ai' PORT=443 TLS=True

    Now, individual options can be accessed or assigned through their
    corresponding attribute:

    >>> settings.PORT
    443
    >>> settings.PORT = 80
    >>> settings.PORT
    80

    .. note::

        Several aggregate representations of options are also available, such as

        >>> settings.dict()
        {'REFRESH_TOKEN': None, 'ACCESS_TOKEN': None, ..., 'TLS': True}
        >>> settings.json()
        '{"REFRESH_TOKEN": null, "ACCESS_TOKEN": null, ..., "TLS": true}'

    Finally, saving a configuration can be done by invoking :meth:`Settings.save`:

    >>> settings.save()
    """

    REFRESH_TOKEN: Optional[str] = None
    """JWT refresh token that can be used to fetch access tokens from the Xanadu Cloud."""

    ACCESS_TOKEN: Optional[str] = None
    """JWT access token that can be used to authenticate requests to the Xanadu Cloud."""

    HOST: str = "platform.xanadu.ai"
    """Hostname of the Xanadu Cloud server."""

    PORT: int = 443
    """Port of the Xanadu Cloud server."""

    TLS: bool = True
    """Whether to use HTTPS for requests to the Xanadu Cloud."""

    class Config:  # pylint: disable=missing-class-docstring
        case_sensitive = True
        env_file = get_path_to_env_file()
        env_prefix = get_name_of_env_var()

    def save(self) -> None:
        """Saves the current settings to the .env file."""

        env_file = Settings.Config.env_file
        env_dir = os.path.dirname(env_file)
        os.makedirs(env_dir, exist_ok=True)

        saved = dotenv_values(dotenv_path=env_file)

        # must be done first as dict is not ordered
        for key, val in self.dict().items():
            _check_for_invalid_values(key, val)

        for key, val in self.dict().items():
            field = get_name_of_env_var(key)

            # Remove keys that are assigned to None.
            if val is None and field in saved:
                unset_key(
                    dotenv_path=env_file,
                    key_to_unset=field,
                    quote_mode="auto",
                )

            # Replace keys that are not assigned to None.
            elif val is not None and val != saved.get(field):
                set_key(
                    dotenv_path=env_file,
                    key_to_set=field,
                    value_to_set=str(val),
                    quote_mode="auto",
                )



# pylint: disable=no-self-use,redefined-outer-name
"""
This module tests the :module:`xcc.job` module.
"""

import io
import json
from datetime import datetime, timedelta
from typing import Callable
from unittest.mock import call, patch

import dateutil.parser
import numpy as np
import numpy.lib.npyio
import pytest
import responses

import xcc


@pytest.fixture
def job_id() -> str:
    """Returns a mock job ID."""
    return "00000000-0000-0000-0000-000000000000"


@pytest.fixture
def job(job_id, connection) -> xcc.Job:
    """Returns a mock job with the given connection."""
    return xcc.Job(job_id, connection)


@pytest.fixture
def datetime_() -> datetime:
    """Returns a mock datetime."""
    return dateutil.parser.isoparse("2020-01-03T12:34:56.789012Z")


@pytest.fixture
def add_response(job_id, connection) -> Callable[[object, str], None]:
    """Returns a function that places a JSON serialization of its argument
    inside the body of an HTTP 200 response to the next GET request to the
    specified path ("/jobs/<job_id>" by default) using the given connection.
    """

    def add_response_(body: object, path: str = f"/jobs/{job_id}") -> None:
        url = connection.url(path)
        responses.add(responses.GET, url, status=200, body=json.dumps(body))

    return add_response_


class TestJob:
    """Tests the :class:`xcc.Job` class."""

    @pytest.mark.parametrize(
        "limit, ids, want_params, want_names",
        [
            (
                1,
                None,
                {"size": "1"},
                ["foo"],
            ),
            (
                2,
                None,
                {"size": "2"},
                ["foo", "bar"],
            ),
            (
                1,
                ["00000000-0000-0000-0000-000000000001"],
                {"size": "1", "id": "00000000-0000-0000-0000-000000000001"},
                ["foo"],
            ),
            (
                2,
                [
                    "00000000-0000-0000-0000-000000000001",
                    "00000000-0000-0000-0000-000000000002",
                    "00000000-0000-0000-0000-000000000003",
                ],
                {
                    "size": "3",
                    "id": [
                        "00000000-0000-0000-0000-000000000001",
                        "00000000-0000-0000-0000-000000000002",
                        "00000000-0000-0000-0000-000000000003",
                    ],
                },
                ["foo", "bar"],
            ),
        ],
    )
    @responses.activate
    def test_list(self, connection, add_response, limit, ids, want_params, want_names):
        """Tests that the correct jobs are listed and that the correct query
        parameters are encoded in the HTTP request to the Xanadu Cloud platform.
        """
        data = [
            {
                "id": "00000000-0000-0000-0000-000000000001",
                "name": "foo",
                "status": "queued",
                "target": "qpu",
                "created_at": "2020-01-03T12:34:56.789012+00:00",
            },
            {
                "id": "00000000-0000-0000-0000-000000000002",
                "name": "bar",
                "status": "complete",
                "target": "qpu",
                "created_at": "2020-01-03T12:34:56.789012+00:00",
                "finished_at": "2020-01-03T12:34:56.789013+00:00",
            },
        ][:limit]
        add_response(body={"data": data}, path="/jobs")

        have_names = [job.name for job in xcc.Job.list(connection, limit=limit, ids=ids)]
        assert have_names == want_names

        have_params = responses.calls[0].request.params  # pyright: reportGeneralTypeIssues=false
        assert have_params == want_params

    @pytest.mark.parametrize(
        "limit, status, want_status_param",
        [
            (1, "queued", "queued"),
            (2, None, None),
            (2, "invalid", "invalid"),
            (2, "complete", "complete"),
        ],
    )
    @responses.activate
    def test_list_status(self, connection, add_response, limit, status, want_status_param):
        """Tests that the correct status parameter is encoded in the HTTP
        request to the Xanadu Cloud platform when listing jobs.
        """
        add_response(body={"data": []}, path="/jobs")
        xcc.Job.list(connection, limit=limit, status=status)

        have_params = responses.calls[0].request.params  # pyright: reportGeneralTypeIssues=false
        have_status_param = have_params.get("status")
        assert have_status_param == want_status_param

    @pytest.mark.parametrize("name", [None, "foo"])
    @responses.activate
    def test_submit(self, job_id, connection, name):
        """Tests that a job can be submitted."""
        details = json.dumps({"id": job_id, "name": name})
        responses.add(responses.POST, connection.url("/jobs"), status=200, body=details)

        job = xcc.Job.submit(connection, name=name, target="bar", circuit="baz", language="qux")
        assert job.id == job_id
        assert job.name == name

    def test_id(self, job):
        """Tests that the correct ID is returned for a job."""
        assert job.id == "00000000-0000-0000-0000-000000000000"

    @responses.activate
    def test_overview(self, job, add_response, datetime_):
        """Tests that the correct overview is returned for a job."""
        body = {
            "id": job.id,
            "name": "foo",
            "status": "complete",
            "target": "bar",
            "language": "baz",
            "created_at": datetime_.isoformat(),
            "finished_at": datetime_.isoformat(),
            "running_time": 0.12345,
            "meta": {},
        }
        add_response(body=body)

        have_keys = set(job.overview)
        want_keys = set(body) ^ {"meta", "metadata"}
        assert have_keys == want_keys

    @responses.activate
    def test_result(self, connection, job):
        """Tests that the correct result is returned for a job."""
        output = [np.complex64(1 + 2j), np.arange(5, dtype=np.int16)]
        metadata = {"params": np.arange(3, dtype=np.int8)}

        with io.BytesIO() as buffer:
            # The public savez() function does not allow pickling to be disabled.
            savez = numpy.lib.npyio._savez  # pylint: disable=protected-access
            savez(
                file=buffer,
                args=output,
                kwds=metadata,
                compress=False,
                allow_pickle=False,
            )

            buffer.seek(0)
            body = buffer.read()

            path = f"/jobs/{job.id}/result"
            responses.add(responses.GET, connection.url(path), status=200, body=body)

            assert job.result.keys() == {"output", "params"}

            assert len(job.result["output"]) == 2
            assert job.result["output"][0] == 1 + 2j
            assert job.result["output"][1] == pytest.approx([0, 1, 2, 3, 4])
            assert job.result["output"][1].dtype == np.int64

            assert job.result["params"] == pytest.approx([0, 1, 2])
            assert job.result["params"].dtype == np.int64

    @responses.activate
    def test_result_npy(self, connection, job):
        """Tests that a TypeError is raised when the result of a job is an .npy file."""
        result = np.array([[0, 1, 2, 3], [4, 5, 6, 7]])

        with io.BytesIO() as buffer:
            np.save(file=buffer, arr=result, allow_pickle=False)

            buffer.seek(0)
            body = buffer.read()

            path = f"/jobs/{job.id}/result"
            responses.add(responses.GET, connection.url(path), status=200, body=body)

            pytest.raises(TypeError, lambda: job.result, match=r"Job result is not an .npz file")

    @responses.activate
    def test_created_at(self, job, add_response, datetime_):
        """Tests that the correct creation time is returned for a job."""
        add_response(body={"started_at": datetime_.isoformat()})
        assert job.started_at == datetime_

    @responses.activate
    def test_started_at_of_unstarted_job(self, job, add_response):
        """Tests that the correct start time is returned for an unstarted job."""
        add_response(body={"started_at": None})
        assert job.started_at is None

    @responses.activate
    def test_started_at_of_started_job(self, job, datetime_, add_response):
        """Tests that the correct start time is returned for a started job."""
        add_response(body={"started_at": datetime_.isoformat()})
        assert job.started_at == datetime_

    @responses.activate
    def test_finished_at_of_unfinished_job(self, job, add_response):
        """Tests that the correct finish time is returned for an unfinished job."""
        add_response(body={"finished_at": None})
        assert job.finished_at is None

    @responses.activate
    def test_finished_at_of_finished_job(self, job, datetime_, add_response):
        """Tests that the correct finish time is returned for a finished job."""
        add_response(body={"finished_at": datetime_.isoformat()})
        assert job.finished_at == datetime_

    @responses.activate
    def test_running_time_of_finished_job(self, job, add_response):
        """Tests that the correct running time is returned for an unfinished job."""
        add_response(body={"running_time": None})
        assert job.running_time is None

    @responses.activate
    def test_running_time_of_unfinished_job(self, job, add_response):
        """Tests that the correct running time is returned for a finished job."""
        add_response(body={"running_time": 12.34})
        assert job.running_time == timedelta(seconds=12.34)

    @responses.activate
    def test_circuit(self, job, add_response):
        """Tests that the correct circuit is returned for a job."""
        add_response(body={"circuit": "foo"}, path=f"/jobs/{job.id}/circuit")
        assert job.circuit == "foo"

    @responses.activate
    def test_language(self, job, add_response):
        """Tests that the correct language is returned for a job."""
        add_response(body={"language": "foo"})
        assert job.language == "foo"

    @pytest.mark.parametrize("name", [None, "foo"])
    @responses.activate
    def test_name(self, job, add_response, name):
        """Tests that the correct name is returned for a job."""
        add_response(body={"name": name})
        assert job.name == name

    @responses.activate
    def test_target(self, job, add_response):
        """Tests that the correct target is returned for a job."""
        add_response(body={"target": "foo"})
        assert job.target == "foo"

    @responses.activate
    def test_status(self, job, add_response):
        """Tests that the correct status is returned for a job."""
        add_response(body={"status": "complete"})
        assert job.status == "complete"

    @pytest.mark.parametrize(
        "status, finished",
        [
            ("open", False),
            ("queued", False),
            ("cancelled", True),
            ("failed", True),
            ("cancel_pending", False),
            ("complete", True),
        ],
    )
    @responses.activate
    def test_finished(self, job, add_response, status, finished):
        """Tests that the correct finished indicator is returned for a job."""
        add_response(body={"status": status})
        assert job.finished is finished

    @responses.activate
    def test_metadata(self, job, add_response):
        """Tests that the correct metadata is returned for a job."""
        add_response(body={"meta": {"foo": "bar", "baz": 123}})
        assert job.metadata == {"foo": "bar", "baz": 123}

    def test_repr(self, job):
        """Tests that the correct printable representation is returned for a job."""
        assert repr(job) == f"<Job: id={job.id}>"

    @responses.activate
    def test_get_result_without_integer_overflow_protection(self, connection, job):
        """Tests that the result of a job can be retrieved without converting
        integral entries into ``np.int64`` objects.
        """
        output = [np.uint8(1), np.int16(2), np.int32(3)]

        with io.BytesIO() as buffer:
            # The public savez() function does not allow pickling to be disabled.
            savez = numpy.lib.npyio._savez  # pylint: disable=protected-access
            savez(file=buffer, args=output, kwds={}, compress=False, allow_pickle=False)

            buffer.seek(0)
            body = buffer.read()

            path = f"/jobs/{job.id}/result"
            responses.add(responses.GET, connection.url(path), status=200, body=body)

            result = job.get_result(integer_overflow_protection=False)

        assert result.keys() == {"output"}
        assert result["output"] == pytest.approx([1, 2, 3])
        assert result["output"][0].dtype == np.uint8
        assert result["output"][1].dtype == np.int16
        assert result["output"][2].dtype == np.int32

    @responses.activate
    def test_cancel(self, connection, job, add_response):
        """Tests that a job can be cancelled."""
        add_response(body={"status": "open"})
        add_response(body={"status": "cancelled"})

        url = connection.url(f"/jobs/{job.id}")
        responses.add(responses.PATCH, url, status=204)

        assert job.finished is False
        assert len(responses.calls) == 1

        job.cancel()
        assert len(responses.calls) == 2

        assert job.finished is True
        assert len(responses.calls) == 3

    @responses.activate
    def test_clear(self, job, add_response):
        """Tests that the cache of a job can be cleared."""
        add_response(body={"status": "queued"})
        add_response(body={"status": "complete"})

        assert job.status == "queued"
        assert len(responses.calls) == 1

        job.clear()
        assert len(responses.calls) == 1

        assert job.status == "complete"
        assert len(responses.calls) == 2

    @responses.activate
    def test_cache_independence(self, connection, job_id, add_response):
        """Tests that caches are not shared across job instances."""
        add_response(body={"status": "open"})
        add_response(body={"status": "queued"})
        add_response(body={"status": "complete"})
        add_response(body={"status": "complete"})

        job_1 = xcc.Job(job_id, connection)
        job_2 = xcc.Job(job_id, connection)

        assert job_1.status == "open"
        assert job_2.status == "queued"

        job_1.clear()

        assert job_1.status == "complete"
        assert job_2.status == "queued"

        job_2.clear()

        assert job_1.status == "complete"
        assert job_2.status == "complete"

    @patch("xcc.job.time.sleep")
    @responses.activate
    def test_wait_on_unfinished_job(self, sleep, job, add_response):
        """Tests that waiting on an unfinished job triggers the correct number
        of polling cycles and that each polling cycle has the correct latency.
        """
        add_response(body={"status": "open"})
        add_response(body={"status": "queued"})
        add_response(body={"status": "complete"})

        job.wait(delay=0.123)
        sleep.assert_has_calls([call(0.123), call(0.123)])

        assert len(responses.calls) == 3
        assert job.finished

    @patch("xcc.job.time.sleep")
    @responses.activate
    def test_wait_on_finished_job(self, sleep, job, add_response):
        """Tests that waiting on a finished job does not trigger a polling
        cycle or send any HTTP requests to the Xanadu Cloud.
        """
        add_response(body={"status": "complete"})

        job.wait(delay=1)
        sleep.assert_not_called()

        assert len(responses.calls) == 1
        assert job.finished



# pylint: disable=no-self-use,redefined-outer-name
"""
This module tests the :module:`xcc.device` module.
"""

import json
from datetime import time
from typing import Callable

import pytest
import responses

import xcc


@pytest.fixture
def device(connection) -> xcc.Device:
    """Returns a mock device with the given connection."""
    return xcc.Device("qpu", connection)


@pytest.fixture
def add_response(connection) -> Callable[[object, str], None]:
    """Returns a function that places a JSON serialization of its argument
    inside the body of an HTTP 200 response to the next GET request to the
    specified path ("/devices/qpu" by default) using the given connection.
    """

    def add_response_(body: object, path: str = "/devices/qpu") -> None:
        url = connection.url(path)
        return responses.add(responses.GET, url, status=200, body=json.dumps(body))

    return add_response_


class TestDevice:
    """Tests the :class:`xcc.Device` class."""

    @pytest.mark.parametrize(
        "status, want_targets",
        [(None, ["foo", "bar"]), ("online", ["foo"]), ("offline", ["bar"])],
    )
    @responses.activate
    def test_list(self, connection, add_response, status, want_targets):
        """Tests that the correct devices are listed for a status."""
        data = [
            {"target": "foo", "status": "online"},
            {"target": "bar", "status": "offline"},
        ]
        add_response(body={"data": data}, path="/devices")

        have_targets = [device.target for device in xcc.Device.list(connection, status)]
        assert have_targets == want_targets

    def test_target(self, device):
        """Tests that the correct target is returned for a device."""
        assert device.target == "qpu"

    @responses.activate
    def test_overview(self, device, add_response):
        """Tests that the correct overview is returned for a device."""
        add_response(body={"target": "qpu", "status": "online"})
        assert device.overview == {"target": "qpu", "status": "online"}

    @responses.activate
    def test_certificate(self, device, add_response):
        """Tests that the correct certificate is returned for a device."""
        add_response(body={"conditions": "nominal"}, path="/devices/qpu/certificate")
        assert device.certificate == {"conditions": "nominal"}

    @responses.activate
    def test_specification(self, device, add_response):
        """Tests that the correct specification is returned for a device."""
        add_response(body={"compiler": "LLVM", "modes": 42}, path="/devices/qpu/specifications")
        assert device.specification == {"compiler": "LLVM", "modes": 42}

    @responses.activate
    def test_expected_uptime(self, device, add_response):
        """Tests that the correct expected uptime is returned for a device."""
        time_up = time.fromisoformat("16:00:00+00:00")
        time_dn = time.fromisoformat("20:00:00+00:00")
        times = [str(time_up), str(time_dn)]

        add_response(body={"expected_uptime": {"wednesday": times}})

        # Take care to verify the iteration order of the expected items.
        have_expected_uptime = list(device.expected_uptime.items())
        want_expected_uptime = [
            ("monday", None),
            ("tuesday", None),
            ("wednesday", (time_up, time_dn)),
            ("thursday", None),
            ("friday", None),
            ("saturday", None),
            ("sunday", None),
        ]

        assert have_expected_uptime == want_expected_uptime

    @responses.activate
    @pytest.mark.parametrize("status", ["offline", "online"])
    def test_status(self, device, add_response, status):
        """Tests that the correct status is returned for a device."""
        add_response(body={"status": status})
        assert device.status == status

    @responses.activate
    @pytest.mark.parametrize("status, up", [("offline", False), ("online", True)])
    def test_up(self, device, add_response, status, up):  # pylint: disable=invalid-name
        """Tests that the correct "up" indicator is returned for a device."""
        add_response(body={"status": status})
        assert device.up == up

    def test_repr(self, device):
        """Tests that the printable representation of a device is correct."""
        assert repr(device) == "<Device: target=qpu>"

    @responses.activate
    def test_clear(self, device, add_response):
        """Tests that the cache of a device can be cleared."""
        add_response(body={"status": "offline"})
        add_response(body={"status": "online"})

        assert device.status == "offline"
        assert device.status == "offline"

        device.clear()

        assert len(responses.calls) == 1
        assert device.status == "online"
        assert len(responses.calls) == 2

    @responses.activate
    def test_cache_independence(self, connection, add_response):
        """Tests that caches are not shared across device instances."""
        add_response(body={"status": "offline"})
        add_response(body={"status": "online"})
        add_response(body={"status": "online"})
        add_response(body={"status": "offline"})

        device_1 = xcc.Device("qpu", connection)
        device_2 = xcc.Device("qpu", connection)

        assert device_1.status == "offline"
        assert device_2.status == "online"
        assert len(responses.calls) == 2

        device_1.clear()

        assert device_1.status == "online"
        assert device_2.status == "online"
        assert len(responses.calls) == 3

        device_2.clear()

        assert device_1.status == "online"
        assert device_2.status == "offline"
        assert len(responses.calls) == 4



"""
This module contains the :class:`~xcc.Device` class.
"""
from __future__ import annotations

import calendar
from datetime import time
from typing import Any, Mapping, Optional, Sequence, Tuple

from .connection import Connection
from .util import cached_property


class Device:
    """Represents a device on the Xanadu Cloud.

    Args:
        target (str): target name of the device
        connection (Connection): connection to the Xanadu Cloud

    .. note::

        For performance reasons, the properties of a device are lazily fetched
        and stored in a cache. This cache can be cleared at any time by calling
        :meth:`Device.clear`.


    .. warning::

        The :class:`xcc.Device` class transparently contacts the Xanadu Cloud
        when an uncached device property is accessed. This means that requesting
        a device property for the first time may take longer than expected.

    **Example:**

    The following example shows how to use the :class:`Device` class to query
    various properties of the ``X8_01`` device on the Xanadu Cloud. First, a
    connection is established to the Xanadu Cloud:

    >>> import xcc
    >>> connection = xcc.Connection.load()

    Next, a reference to the ``X8_01`` device is created using the connection.

    >>> device = xcc.Device(target="X8_01", connection=connection)

    Finally, the certificate, specification, status, etc. of the ``X8_01``
    device are retrieved by accessing the corresponding property of the device:

    >>> device.certificate
    {'device_url': ..., 'laser_wavelength_meters': 1.55270048e-06}
    >>> device.specification
    {'gate_parameters': ..., 'target': 'X8_01'}
    >>> device.status
    'online'
    """

    @staticmethod
    def list(connection: Connection, status: Optional[str] = None) -> Sequence[Device]:
        """Returns devices on the Xanadu Cloud.

        Args:
            connection (Connection): connection to the Xanadu Cloud
            status (str, optional): optionally filter devices by status

        Returns:
            Sequence[Device]: devices on the Xanadu Cloud which match the status filter
        """
        response = connection.request("GET", "/devices")

        def include(details: Mapping[str, Any]) -> bool:
            """Returns ``True`` if a device with the given details should be
            included in the response.  Otherwise, ``False`` is returned.
            """
            return status is None or details["status"] == status

        devices = []

        for details in filter(include, response.json()["data"]):
            device = Device(target=details["target"], connection=connection)
            device._details = details  # pylint: disable=protected-access
            devices.append(device)

        return devices

    def __init__(self, target: str, connection: Connection) -> None:
        self._target = target
        self._connection = connection

    @property
    def target(self) -> str:
        """Returns the target of a device."""
        return self._target

    @property
    def overview(self) -> Mapping[str, Any]:
        """Returns an overview of a device.

        Returns:
            Mapping[str, Any]: mapping from field names to values for this
            device as determined by the needs of a Xanadu Cloud user.
        """
        return {"target": self.target, "status": self.status}

    @cached_property
    def certificate(self) -> Mapping[str, Any]:
        """Returns the certificate of a device.

        A device certificate contains the current operating conditions of a
        device and is periodically updated while the device is online.

        Returns:
            Mapping[str, Any]: certificate of this device

        .. warning::

            The structure of a certificate may vary from device to device.
        """
        return self._connection.request("GET", f"/devices/{self.target}/certificate").json()

    @cached_property
    def specification(self) -> Mapping[str, Any]:
        """Returns the specification of a device.

        Returns:
            Mapping[str, Any]: specification of this device
        """
        return self._connection.request("GET", f"/devices/{self.target}/specifications").json()

    @property
    def expected_uptime(self) -> Mapping[str, Optional[Tuple[time, time]]]:
        """Returns the expected uptime of a device.

        Returns:
            Mapping[str, Optional[Tuple[time, time]]]: mapping from weekdays to
            optional time pairs where each pair represents when the device is
            expected to come online and offline that day
        """
        # Ensure that the iteration order is chronological.
        expected_uptime = dict.fromkeys(map(str.lower, calendar.day_name))

        for weekday, times in self._details["expected_uptime"].items():
            expected_uptime[weekday] = tuple(map(time.fromisoformat, times))

        return expected_uptime

    @property
    def status(self) -> str:
        """Returns the status of a device.

        Returns:
            str: status of this device ("online" or "offline")
        """
        return self._details["status"]

    @property
    def up(self) -> bool:  # pylint: disable=invalid-name
        """Returns whether a device is accepting jobs.

        Returns:
            bool: ``True`` iff the status of this device is "online"
        """
        return self.status == "online"

    @cached_property
    def _details(self) -> Mapping[str, Any]:
        """Returns the details of a device.

        Returns:
            Mapping[str, Any]: mapping from field names to values for this
            device as determined by the Xanadu Cloud device endpoint.

        .. note::

            These fields are not intended to be directly accessed by external
            callers. Instead, they should be individually retrieved through
            their associated public properties.
        """
        return self._connection.request("GET", f"/devices/{self.target}").json()

    def __repr__(self) -> str:
        """Returns a printable representation of a device."""
        return f"<{self.__class__.__name__}: target={self.target}>"

    def clear(self) -> None:
        """Clears the details, certificate, and specification caches of a device."""
        del self._details
        del self.certificate
        del self.specification



"""
This module implements the Xanadu Cloud CLI.
"""

import json
import sys
from functools import wraps
from pprint import pformat
from typing import Any, Callable, List, Mapping, Sequence, Tuple, Union

import fire
import numpy as np
from fire.core import FireError
from fire.formatting import Error
from pydantic import ValidationError

from ._version import __version__
from .connection import Connection
from .device import Device
from .job import Job
from .settings import Settings


def beautify(command: Callable) -> Callable:
    """Decorator which formats the output of a CLI command.

    Args:
        command (Callable): function which implements a CLI command

    Returns:
        Callable: CLI command which converts the value returned by the given
        CLI command into a JSON string if the value is a list or dictionary
    """

    @wraps(command)
    def beautify_(*args, **kwargs):
        output = command(*args, **kwargs)
        if isinstance(output, (list, dict)):
            return json.dumps(output, indent=4, default=str)
        return output

    return beautify_


def load_connection() -> Connection:
    """Loads a connection using :func:`xcc.Connection.load()` with a "User-Agent"
    header that specifies the Xanadu Cloud Client CLI.

    Returns:
        Connection: connection with a "User-Agent" header that is appropriate
        for the CLI, loaded from a new :class:`xcc.Settings` instance

    .. warning::

        This is an internal XCC utility function and should *not* be called from
        outside the XCC package. Use :func:`xcc.Connection.load()` instead.
    """
    return Connection.load(headers={"User-Agent": f"XCC/{__version__} (CLI)"})


# Settings CLI
# ------------------------------------------------------------------------------


@beautify
def get_setting(name: str) -> Any:
    """Gets the value of a setting.

    Args:
        name (str): Name of the setting (e.g., "PORT").

    Returns:
        Any: Value of the setting.
    """
    key, val = _resolve_setting(name)

    if val is None:
        raise ValueError(f"The {key} setting does not currently have a value.")

    return val


@beautify
def list_settings() -> Mapping[str, Any]:
    """Lists the current settings.

    Returns:
        Mapping[str, Any]: Mapping from setting names to values.
    """
    return Settings().dict()


@beautify
def set_setting(name: str, value: Union[str, int, bool]) -> str:
    """Sets the value of a setting.

    To configure your Xanadu Cloud credentials, simply run

    $ xcc config set REFRESH_TOKEN "Xanadu Cloud API key goes here"

    Args:
        name (str): Name of the setting (e.g., "PORT").
        value (str, int, bool): Value of the setting (e.g., 443).

    Returns:
        str: Message indicating the setting was successfully updated.
    """
    key, _ = _resolve_setting(name)

    try:
        settings = Settings(**{key: value})
        settings.save()
    except ValidationError as exc:
        err = exc.errors()[0].get("msg", "invalid value")
        raise ValueError(f"Failed to update {key} setting: {err}") from exc

    # Using repr() ensures that strings are quoted.
    val = repr(settings.dict()[key])
    return f"Successfully updated {key} setting to {val}."


def _resolve_setting(name: str) -> Tuple[str, Any]:
    """Resolves the key and value of a setting.

    Args:
        name (str): name of the setting

    Returns:
        Tuple[str, Any]: resolved key and value of the setting

    Raises:
        ValueError: if the name of the setting is invalid
    """
    key = name.upper()

    settings = Settings()
    if key not in settings.dict():
        raise ValueError(f"The setting name '{name}' must be one of {list(settings.dict())}.")

    return key, settings.dict()[key]


# Device CLI
# ------------------------------------------------------------------------------


@beautify
def get_device(
    target: str,
    availability: bool = False,
    certificate: bool = False,
    specification: bool = False,
    status: bool = False,
) -> Union[str, Mapping]:
    """Gets information about a device on the Xanadu Cloud.

    If no device property flags are specified, an overview of the device is
    displayed. Otherwise, only the selected device property is shown.

    Args:
        target (str): Name of the device target.
        availability (bool): Show the expected uptime of the device.
        certificate (bool): Show the certificate of the device.
        specification (bool): Show the specification of the device.
        status (bool): Show the status of the device.

    Returns:
        Union[str, Mapping]: Device overview or the selected device property.
    """
    device = Device(target=target, connection=load_connection())

    flags = sum(map(int, (availability, certificate, specification, status)))
    if flags > 1:
        raise FireError("At most one device property can be selected.")

    if availability:
        return device.expected_uptime
    if certificate:
        return device.certificate
    if specification:
        return device.specification
    if status:
        return device.status

    return device.overview


@beautify
def list_devices(status: str = None) -> Sequence[Mapping]:
    """Lists devices on the Xanadu Cloud.

    Args:
        status (str): Filter devices by status (e.g., "offline" or "online").

    Returns:
        Sequence[Mapping]: Overview of each device on the Xanadu Cloud.
    """
    devices = Device.list(connection=load_connection(), status=status)
    return [device.overview for device in devices]


# Job CLI
# ------------------------------------------------------------------------------


@beautify
# pylint: disable=invalid-name,redefined-builtin
def cancel_job(id: str) -> str:
    """Cancels a job on the Xanadu Cloud.

    Args:
        id (str): ID of the job.

    Returns:
        str: Message indicating the job was successfully cancelled.
    """
    job = Job(id_=id, connection=load_connection())
    job.cancel()

    if job.status not in ("cancel_pending", "cancelled"):
        raise ValueError(f"Job with ID '{id}' was not cancelled in time.")

    return f"Successfully cancelled job with ID '{id}'."


@beautify
# pylint: disable=invalid-name,redefined-builtin
def get_job(
    id: str, circuit: bool = False, result: bool = False, status: bool = False
) -> Union[str, Mapping]:
    """Gets information about a job on the Xanadu Cloud.

    If no job property flags are specified, an overview of the job is
    displayed. Otherwise, only the selected job property is shown.

    Args:
        id (str): ID of the job.
        circuit (bool): Show the circuit of the job.
        result (bool): Show the result of the job.
        status (bool): Show the status of the job.

    Returns:
        Union[str, Mapping]: Job overview or the selected job property.
    """
    job = Job(id_=id, connection=load_connection())

    flags = sum(map(int, (circuit, result, status)))
    if flags > 1:
        raise FireError("At most one job property can be selected.")

    if circuit:
        return job.circuit
    if status:
        return job.status
    if result:
        result = job.get_result(integer_overflow_protection=False)

        def convert_to_list(value: Union[np.ndarray, List[np.ndarray]]) -> List:
            """Converts the given job result value into a native Python list."""
            if isinstance(value, np.ndarray):
                return value.tolist()

            return [array.tolist() for array in value]

        # Convert all NumPy arrays into lists to take advantage of pprint.
        list_result = {key: convert_to_list(value) for key, value in result.items()}

        # Use "" instead of '' to ensure the output is compatible with JSON.
        json_result = pformat(list_result).replace("'", '"')
        # Apply the same formatting rules as the JSON dumper in beautify().
        return json_result.replace("{", "{\n ").replace("\n", "\n   ").replace("}", "\n}")

    return job.overview


@beautify
def list_jobs(limit: int = 5, ids: List[str] = None, status: str = None) -> Sequence[Mapping]:
    """Lists jobs submitted to the Xanadu Cloud.

    Args:
        limit (int): Maximum number of jobs to display.
        ids (List[str], optional): IDs of the jobs to display. If at least one
            ID is specified, the limit flag will be set to the number of IDs.
        status (str, optional): If specificed, filter returned jobs by the given status.

    Returns:
        Sequence[Mapping]: Overview of each job submitted to the Xanadu Cloud.
    """
    jobs = Job.list(connection=load_connection(), limit=limit, ids=ids, status=status)
    return [job.overview for job in jobs]


@beautify
def submit_job(
    target: str, circuit: str, language: str = "blackbird:1.0", name: str = None
) -> Mapping:
    """Submits a job to the Xanadu Cloud.

    Args:
        target (str): Target of the job.
        circuit (str): Circuit of the job.
        language (str): Language of the job.
        name (str, optional): Name of the job.

    Returns:
        Mapping: Overview of the submitted job.
    """
    job = Job.submit(
        connection=load_connection(),
        name=name,
        target=target,
        circuit=circuit.replace("\\n", "\n"),
        language=language,
    )
    return job.overview


# Other CLI
# ------------------------------------------------------------------------------


def ping() -> str:
    """Tests the connection to the Xanadu Cloud.

    Returns:
        str: Message indicating the ping was successful.
    """
    connection = load_connection()
    connection.ping()
    return "Successfully connected to the Xanadu Cloud."


def version() -> str:
    """Displays the version of the Xanadu Cloud Client.

    Returns:
        str: Message with the version of this Python package.
    """
    return f"Xanadu Cloud Client version {__version__}"


def main() -> None:
    """Entry point for the Xanadu Cloud CLI."""
    # Using a pager to display help information can be annoying since it hides
    # the output (and possible error messages) of previous commands. To remedy
    # this, the assignment below replaces the pager with the print() function.
    # See https://github.com/google/python-fire/issues/188 for more details.
    fire.core.Display = lambda lines, out: print(
        *lines, file=out
    )  # pyright: reportGeneralTypeIssues=false

    try:
        fire.Fire(
            {
                "config": {
                    "get": get_setting,
                    "list": list_settings,
                    "set": set_setting,
                },
                "device": {
                    "get": get_device,
                    "list": list_devices,
                },
                "job": {
                    "cancel": cancel_job,
                    "get": get_job,
                    "list": list_jobs,
                    "submit": submit_job,
                },
                "ping": ping,
                "version": version,
            }
        )

    # pylint: disable=broad-except
    except Exception as exc:
        # Some exceptions, like NotImplementedError, don't have a message.
        err = str(exc).rstrip(".") or exc.__class__.__name__
        msg = Error("ERROR: ") + err + "."
        print(msg, file=sys.stderr)
